/***************************************************************************
 *   Copyright (C) 2006 by Alexios Beveratos   *
 *   alexios@beveratos.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
// Some code has been extracted from gpscorrelate by Daniel Foote
/* exif-gps.cpp
 * Written by Daniel Foote.
 * Started Feb 2005.
 *
 * This file contains routines for reading dates
 * from exif data, and writing GPS data into the
 * appropriate photos.
 */

/* Copyright 2005 Daniel Foote.
 *
 * This file is part of gpscorrelate.
 */
 
#include "exif_gps.h"
#include <qdatetime.h>
#include <kfileitem.h>
#include <qstring.h>
#include <math.h>
#include <iostream>
#include <iomanip>

#include <exiv2/image.hpp>
#include <exiv2/exif.hpp>
#include <exiv2/error.hpp>

Exif_GPS::Exif_GPS()
{
}


Exif_GPS::~Exif_GPS()
{
}


void Exif_GPS::ConvertToRational(double Number,
		       long int* Numerator, long int* Denominator,
		       int Rounding)
{
	// This function converts the given decimal number
	// to a rational (fractional) number.
	//
	// Examples in comments use Number as 25.12345, Rounding as 4.
	
	// Split up the number.
	double Whole = trunc(Number);
	double Fractional = Number - Whole;

	// Calculate the "number" used for rounding.
	// This is 10^Digits - ie, 4 places gives us 10000.
	double Rounder = pow(10, Rounding);

	// Round the fractional part, and leave the number
	// as greater than 1.
	// To do this we: (for example)
	//  0.12345 * 10000 = 1234.5
	//  floor(1234.5) = 1234 - now bigger than 1 - ready...
	Fractional = trunc(Fractional * Rounder);

	// Convert the whole thing to a fraction.
	// Fraction is:
	//     (25 * 10000) + 1234   251234
	//     ------------------- = ------ = 25.1234
	//           10000            10000
	double NumTemp = (Whole * Rounder) + Fractional;
	double DenTemp = Rounder;

	// Now we should reduce until we can reduce no more.
	
	// Try simple reduction...
	// if   Num
	//     ----- = integer out then....
	//      Den
	if (trunc(NumTemp / DenTemp) == (NumTemp / DenTemp))
	{
		// Divide both by Denominator.
		NumTemp /= DenTemp;
		DenTemp /= DenTemp;
	}
	
	// And, if that fails, brute force it.
	while (1)
	{
		// Jump out if we can't integer divide one.
		if ((NumTemp / 2) != trunc(NumTemp / 2)) break;
		if ((DenTemp / 2) != trunc(DenTemp / 2)) break;
		// Otherwise, divide away.
		NumTemp /= 2;
		DenTemp /= 2;
	}

	// Copy out the numbers.
	*Numerator = (int)NumTemp;
	*Denominator = (int)DenTemp;

	// And finished...

}

QDateTime Exif_GPS::extractDateTime(const KFileItem* selectedFile)
{
	Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( (selectedFile->url()).path().ascii()  );
	image->readMetadata();
	Exiv2::ExifData &exifData = image->exifData();

	Exiv2::ExifData::const_iterator i = exifData.begin();

	i=exifData.findKey( Exiv2::ExifKey::ExifKey  ( "Exif.Image.DateTime" ) );

	QString raw = QString( i->value().toString () );
	QString DateString = raw.section(" ",0,0).stripWhiteSpace();
	QString TimeString = raw.section(" ",1,1).stripWhiteSpace();
	
	QString StringDateTime=DateString.section(":",0,0)+"-"+
			DateString.section(":",1,1)+"-"+DateString.section(":",2,2);
	StringDateTime+="T"+TimeString.section(":",0,0)+":"+
			TimeString.section(":",1,1)+":"+TimeString.section(":",2,2);

	return QDateTime::fromString( StringDateTime, Qt::ISODate);

}

//bool Exif_GPS::tagPhoto( double Alt, double Lat, double Long, const KFileItem* selectedFile)
bool Exif_GPS::tagPhoto( double Alt, double Lat, double Long, QString selectedFile)
{
	Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( selectedFile.ascii()  );
	image->readMetadata();
	
	Exiv2::ExifData &ExifToWrite = image->exifData();
	
	char ScratchBuf[100];
	long int Nom, Denom;
	long int Deg, Min;
	
	// Do all the easy constant ones first.
	// GPSVersionID tag: standard says is should be four bytes: 02 00 00 00
	//  (and, must be present).
	Exiv2::Value::AutoPtr Value = Exiv2::Value::create(Exiv2::unsignedByte);
	Value->read("2 0 0 0");
	
	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSVersionID"), Value.get());
	// Datum: the datum of the measured data. If not given, we insert WGS-84.
	//ExifToWrite["Exif.GPSInfo.GPSMapDatum"] = Datum;
	ExifToWrite["Exif.GPSInfo.GPSMapDatum"] = "WGS-84";
	
	// Now start adding data.
	// ALTITUDE.
	// Altitude reference: byte "00" meaning "sea level".
	
	Value = Exiv2::Value::create(Exiv2::unsignedByte);
	Value->read("0");
	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSAltitudeRef"), Value.get());
	// And the actual altitude.
	Value = Exiv2::Value::create(Exiv2::signedRational);
	Exif_GPS::ConvertToRational(Alt, &Nom, &Denom, 4);
	snprintf(ScratchBuf, 100, "%ld/%ld", Nom, Denom);
	Value->read(ScratchBuf);
	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSAltitude"), Value.get());
	
	// LATTITUDE
	// Latitude reference: "N" or "S".
	if (Lat < 0)
	{
		// Less than Zero: ie, minus: means
		// Southern hemisphere. Where I live.
		ExifToWrite["Exif.GPSInfo.GPSLatitudeRef"] = "S";
	} else {
		// More than Zero: ie, plus: means
		// Northern hemisphere.
		ExifToWrite["Exif.GPSInfo.GPSLatitudeRef"] = "N";
	}
	// Now the actual lattitude itself.
	// This is done as three rationals.
	// I choose to do it as:
	//   dd/1 - degrees.
	//   mmmm/100 - minutes
	//   0/1 - seconds
	// Exif standard says you can do it with minutes
	// as mm/1 and then seconds as ss/1, but its
	// (slightly) more accurate to do it as
	//  mmmm/100 than to split it.
	// We also absolute the value (with fabs())
	// as the sign is encoded in LatRef.
	// Further note: original code did not translate between
	//   dd.dddddd to dd mm.mm - that's why we now multiply
	//   by 6000 - x60 to get minutes, x100 to get to mmmm/100.
	Value = Exiv2::Value::create(Exiv2::signedRational);
	Deg = (int)floor(fabs(Lat)); // Slice off after decimal.
	Min = (int)floor((fabs(Lat) - floor(fabs(Lat))) * 6000);
	snprintf(ScratchBuf, 100, "%ld/1 %ld/100 0/1", Deg, Min);
	Value->read(ScratchBuf);
	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSLatitude"), Value.get());
	
	// LONGITUDE
	// Longitude reference: "E" or "W".
	if (Long < 0)
	{
		// Less than Zero: ie, minus: means
		// Western hemisphere.
		ExifToWrite["Exif.GPSInfo.GPSLongitudeRef"] = "W";
	} else {
		// More than Zero: ie, plus: means
		// Eastern hemisphere. Where I live.
		ExifToWrite["Exif.GPSInfo.GPSLongitudeRef"] = "E";
	}
	// Now the actual longitude itself.
	// This is done as three rationals.
	// I choose to do it as:
	//   dd/1 - degrees.
	//   mmmm/100 - minutes
	//   0/1 - seconds
	// Exif standard says you can do it with minutes
	// as mm/1 and then seconds as ss/1, but its
	// (slightly) more accurate to do it as
	//  mmmm/100 than to split it.
	// We also absolute the value (with fabs())
	// as the sign is encoded in LongRef.
	// Further note: original code did not translate between
	//   dd.dddddd to dd mm.mm - that's why we now multiply
	//   by 6000 - x60 to get minutes, x100 to get to mmmm/100.
	Value = Exiv2::Value::create(Exiv2::signedRational);
	Deg = (int)floor(fabs(Long)); // Slice off after decimal.
	Min = (int)floor((fabs(Long) - floor(fabs(Long))) * 6000);
	snprintf(ScratchBuf, 100, "%ld/1 %ld/100 0/1", Deg, Min);
	Value->read(ScratchBuf);
	ExifToWrite.add(Exiv2::ExifKey("Exif.GPSInfo.GPSLongitude"), Value.get());

	// Write the data to file.
	image->writeMetadata();

	return TRUE;
	
}
